import 'dart:async';
import 'dart:convert';

import 'package:http/http.dart';
import 'package:puma/src/args/args.dart';
import 'package:puma/src/args/global.dart';
import 'package:puma/src/command/flutter.dart';
import 'package:puma/src/command/help.dart';
import 'package:puma/src/command/package.dart';
import 'package:puma/src/command/version.dart';
import 'package:puma/src/pubspec.dart';
import 'package:puma/src/command/errors.dart';
import 'package:puma/src/commands.dart';
import 'package:puma/src/puma.dart';
import 'package:puma/src/utils/catch_show_usage_exception.dart';
import 'package:puma/src/utils/get_command.dart';
import 'package:puma/src/utils/write_content.dart';
import 'package:yaml/yaml.dart' as yaml;
import 'package:args/args.dart';

class PumaCommander {
  final String content;
  final Client client;
  final GetLocalPackage getLocalPackage;
  final GetAssetsList getAssetsList;

  final _newContentController = StreamController<String>();
  final _helpController = StreamController<String>();
  final _needPubGet = StreamController<bool>();

  PumaCommander(
    this.content,
    this.client,
    this.getLocalPackage,
    this.getAssetsList,
  );

  Stream<String> get pubspec => _newContentController.stream;

  Stream<String> get help => _helpController.stream;

  Stream<bool> get needPubGet => _needPubGet.stream;

  Future<void> init(List<String> args) async {
    try {
      final results = parser.parse(args);
      final globalArgs = parseGlobalArgs(args);
      if (args.contains('--pub-host-url')) {
        args..removeRange(args.indexOf('--pub-host-url'), 2);
      }
      final allCommands = [
        PackageCommand(args),
        VersionCommand(args),
        FlutterCommand(args),
        HelpCommand([], args),
      ];
      final helpCommand = HelpCommand(allCommands, args);
      try {
        final yamlRoot = yaml.loadYaml(content) as yaml.YamlMap;
        final oldPubspec = Pubspec.fromJson(
          (json.decode(json.encode(yamlRoot)) as Map).cast<String, dynamic>(),
        );
        final puma = Puma(
          globalArgs.pubHostUrl,
          client,
          oldPubspec,
          getLocalPackage,
          getAssetsList,
        );
        final commands = Commands(
          allCommands,
          helpCommand,
        );
        final pubspec = await commands.run(results, puma);
        if (pubspec != null) {
          if (pubspec.needPubGet(oldPubspec)) {
            _needPubGet.add(true);
          }
          final pubspecContent = await concatContent(pubspec.toYaml());
          _newContentController.add(pubspecContent);
        }
      } on ShowUsageException catch (e) {
        _helpController.add(e.usage);
      } catch (e, s) {
        final sb = StringBuffer();
        sb.writeln('Oops you found a bug in puma !');
        sb.writeln();
        sb.writeln(
            'Please report us the error here : https://gitlab.com/santetis/open_source/puma/issues/new');
        sb.writeln(
            'If you can include this message it will save us lot of time :)');
        sb.writeln();
        sb.writeln(e);
        sb.writeln(s);
        _helpController.add(sb.toString());
      }
    } on ArgParserException catch (e) {
      final allCommands = [
        PackageCommand(args),
        VersionCommand(args),
        FlutterCommand(args),
        HelpCommand([], args),
      ];
      final helpCommand = HelpCommand(allCommands, args);
      final command = await getCommand(e.commands, allCommands) ?? helpCommand;
      final help = await catchShowUsageException(() async {
        await command.run(null, null);
      });
      if (help != null) {
        if (e.commands.isNotEmpty) {
          _helpController.add(help);
        } else {
          _helpController.add('${e.message}\n\n$help');
        }
      }
    } finally {
      _newContentController.close();
      _helpController.close();
      _needPubGet.close();
    }
  }
}
